/*
MIT License

Copyright (c) 2021 Famedly

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

import 'dart:async';
import 'dart:convert';

import 'package:http/http.dart';

import 'json_rest_api_exception.dart';
import 'json_rest_api_response.dart';

/// Base class to create a json REST API client.
class JsonRestApiClient {
  /// The base URL which will be the prefix for all requests.
  final Uri baseUrl;

  /// If set, all requests will send this token in the Authorization header
  /// with `Bearer $bearerToken`.
  String? bearerToken;

  int _timeoutFactor = 1;

  final Client _httpClient;

  /// Set the [baseUrl] and an optional [bearerToken]. Overwrite [httpClient]
  /// with a mock client for testing.
  JsonRestApiClient(
    this.baseUrl, {
    this.bearerToken,
    Client? httpClient,
  }) : _httpClient = httpClient ?? Client();

  /// Perform a request to the API which could either return json data or null.
  /// Set [method] to a valid http method like `GET` or `POST` and endpoint to
  /// one of the available endpoints of your api like `/endpoint`.
  ///
  /// Send some data using [json] (can be Map or List) or use the
  /// [queryParamters].
  ///
  /// You can also set a custom [timeout] duration or add custom [headers].
  Future<JsonRestApiResponse> request(
    String method,
    String endpoint, {
    dynamic? json,
    Map<String, dynamic>? queryParameters,
    Duration? timeout,
    Map<String, String> headers = const {},
  }) async {
    // Calculate final url
    final url = baseUrl.resolveUri(
      Uri(
        path: endpoint,
        queryParameters: queryParameters,
      ),
    );

    // Prepare request
    final request = Request(method, url);
    if (json != null) {
      request.body = jsonEncode(json);
    }

    // Set headers
    if (['PUT', 'POST', 'PATCH'].contains(method)) {
      request.headers['Content-Type'] = 'application/json';
    }
    if (bearerToken != null) {
      request.headers['Authorization'] = 'Bearer $bearerToken';
    }
    for (final entry in headers.entries) {
      request.headers[entry.key] = entry.value;
    }

    try {
      // Start the request
      final response = await _httpClient.send(request).timeout(
            timeout ?? Duration(seconds: _timeoutFactor + 5),
          );
      var responseBody = await response.stream.bytesToString();

      // Put arrays into a wrapper map
      List<dynamic>? listResp;
      Map<String, dynamic>? jsonResp;
      var jsonString = String.fromCharCodes(responseBody.runes);
      if (jsonString.isNotEmpty) {
        try {
          final json = jsonDecode(jsonString);
          if (json is Map<String, dynamic>) {
            jsonResp = json;
          } else if (json is List) {
            listResp = json;
          }
        } on FormatException catch (_) {}
      }

      _timeoutFactor = 1;
      final jsonResponse = JsonRestApiResponse(
        endpoint: endpoint,
        method: method,
        raw: jsonString,
        statusCode: response.statusCode,
        jsonObject: jsonResp,
        jsonArray: listResp,
      );

      if (response.statusCode >= 300 || response.statusCode < 200) {
        throw JsonRestApiException.fromJsonRestApiResponse(
          jsonResponse,
          reasonPhrase: response.reasonPhrase,
        );
      }
      return jsonResponse;
    } on TimeoutException catch (_) {
      _timeoutFactor *= 2;
      rethrow;
    }
  }
}
